/* regmon 01.08.2012 */

#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include <unistd.h>
#include <termios.h>
#ifndef WIN32
#include <sys/mman.h>
#include <fcntl.h>
#endif

#define MAP_SIZE 4096UL
#define MAP_MASK (MAP_SIZE - 1)
#define CMD_LEN_MAX 80

typedef enum {
    CMD_UNKNOWN,
    CMD_GETMEM,
    CMD_SETMEM,
    CMD_BYTE,
    CMD_WORD,
    CMD_DWORD,
    CMD_COL,
    CMD_HELP,
    CMD_LITTLE_E,
    CMD_BIG_E,
    CMD_QUIT
} Cmd;

typedef struct
{
    Cmd cmd;
    char* cmdStr;
} cmdEntry;

cmdEntry cmdTable[] =
{
    {CMD_GETMEM,    "md"},
    {CMD_SETMEM,    "mm"},
    {CMD_BYTE,      "b"},
    {CMD_WORD,      "w"},
    {CMD_DWORD,     "d"},
    {CMD_COL,       "c"},
    {CMD_QUIT,      "q"},
    {CMD_LITTLE_E,  "le"},
    {CMD_BIG_E,     "be"},
    {CMD_HELP,      "?"},
    {CMD_HELP,      "h"},
    {CMD_HELP,      "help"}
};

unsigned int columns=1;
unsigned int size=4;
unsigned int addr;
unsigned int len;
unsigned char littleEndian=1;
int mem;

#ifdef TEST_MODE
char testData[1024];
#endif

void usage() {
    printf("Display/modify registers at physical addresses\n");
    printf("Usage:\n");
    printf("md <addr> <count>: Memory dump from <addr>; <count> elements\n");
    printf("mm <addr>        : Memory modify at <addr> with auto-increment\n");
    printf("                   Enter non-hex number to exit\n");
    printf("d                : display data as DWORDs (default)\n");
    printf("w                : display data as WORDs\n");
    printf("b                : display data as BYTEs\n");
    printf("c <n>            : display data in <n> columns (default=1)\n");
    printf("le               : little endian mode (default)\n");
    printf("be               : big endian mode \n");
    printf("?                : guess what?\n");
    printf("q                : quit\n\n");
}


char getch() {
	char buf = 0;
	struct termios old = {0};
	if (tcgetattr(0, &old) < 0)
		perror("tcsetattr()");
	old.c_lflag &= ~ICANON;
	old.c_lflag &= ~ECHO;
	old.c_cc[VMIN] = 1;
	old.c_cc[VTIME] = 0;
	if (tcsetattr(0, TCSANOW, &old) < 0)
		perror("tcsetattr ICANON");
	if (read(0, &buf, 1) < 0)
		perror ("read()");
	old.c_lflag |= ICANON;
	old.c_lflag |= ECHO;
	if (tcsetattr(0, TCSADRAIN, &old) < 0)
		perror ("tcsetattr ~ICANON");
	return (buf);
}

void readCmdStr(char *rawCmdstr, unsigned int maxLen)
{
	char c;
	char *p = rawCmdstr;
/*	int n; */

	fflush(stdout);
	do {
		c = getch();
		switch (c) {
		case 0x08:
		case 0x7f:
			if (p>rawCmdstr) {
				p--;
				printf("%c",0x08);
				printf("%c",' ');
				printf("%c",0x08);
			}
			break;
		case 0x0a:
			break;
		default:
#if 1
		// check for ESC-Sequences
			if (c == 0x1b) {
				c = getch();
				// check for cursor keys
				if (c == 0x5b) {
					c = getch();
					// check for PgUp & Co.
					if ((c >= 0x31) && (c <= 0x36)){
						c = getch();
						break;
					}
					break;
				}
			}
#endif
			*p++ = c;
			putchar(c);
			break;
		}
		fflush(stdout);
	} while (c != 0x0a);
	*p = 0;
#if 0
	for(n=0; n<=strlen(rawCmdstr); n++)
		printf("\n%02x ", rawCmdstr[n]);
#endif
}


Cmd getCmd(char** cmdstr)
{
    int n;
    char main_cmd[CMD_LEN_MAX];

    /* printf("Processing Input: %s\n", *cmdstr); */
    if (sscanf(*cmdstr, "%s", main_cmd) < 1) {
        return CMD_UNKNOWN;
    }
    /* printf("Parsed Input: %s\n", main_cmd);
    printf("Searching %d entries\n", sizeof(cmdTable)/sizeof(cmdEntry)); */
    for (n=0; n<(sizeof(cmdTable)/sizeof(cmdEntry));n++)
    {
        /* printf("comparing %s with %s\n", main_cmd, cmdTable[n].cmdStr); */
        if (!strcmp(main_cmd, cmdTable[n].cmdStr)) {
            *cmdstr += strlen(cmdTable[n].cmdStr);
            return cmdTable[n].cmd;
        }
    }
    return CMD_UNKNOWN;
}


void dump(char* args) {
    unsigned int n, c, l;       /* counter, column, length */
    char* pa;                   /* physical address */
    char* va;                   /* virtual address */
    void *map_base;

    /* printf("Dumping: %s\n", args); */

    switch (sscanf(args, "%x %x", &pa, &l)) {
        case 1:
            l = 1;
            break;
        case 2:
            break;
        default:
            printf("Argument missing or wrong\n");
            return;
    }

#ifdef TEST_MODE
    // use prepared data
    va = (unsigned char*)testData;
#endif

    c = columns;

    while(l) {
#ifndef TEST_MODE
    // Map the device to memory
        map_base = mmap(0, MAP_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, mem, (off_t)pa & ~MAP_MASK);
        va = map_base + ((off_t)pa & MAP_MASK);
#endif
        if (c == columns) {
            printf("%08X: ", pa);
            c = 0;
        }
        while (l && (c<columns)) {
            for (n=0; n<size; n++) {
                /* method to compensate different endianess' offsets:
                   byte offset      BE          LE      delta
                    n=0             +0          +3      +3
                    n=1             +1          +2      +1
                    n=2             +2          +1      +1
                    n=3             +3          +0      -3

                ==> offset for BE (LE=0)   : va+n
                plus delta in case of LE==1: (size-1)-2n */
                printf("%02X", *(va+n+littleEndian*((size-1)-(n<<1))));
            }
            printf(" ");
            va += size;
            pa += size;
            c++;
            l--;
        }
        printf("\n");
#ifndef TEST_MODE
        munmap(map_base, MAP_SIZE);
#endif
    };
}

void modify(char* args) {
    unsigned int d;             /* counter, data */
    unsigned long* pa;          /* physical address */
    unsigned long* va;          /* virtual address */
    char dataStr[CMD_LEN_MAX];
    unsigned int busy = 1;
    void *map_base;

    if (sscanf(args, "%lx", &pa) < 1) {
        printf("Argument missing or wrong\n");
        return;
    }

#ifdef TEST_MODE
    // use prepared data
    va = (unsigned int*)testData;
#endif

    do {
#ifndef TEST_MODE
        // Map the device to memory
        map_base = mmap(0, MAP_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, mem, (off_t)pa & ~MAP_MASK);
        va = map_base + ((off_t)pa & MAP_MASK);
#endif
        printf("%08X: [%08X] ", pa, *va);

        if (fgets(dataStr, CMD_LEN_MAX, stdin) == NULL) {
            pa++;
        } else {

            if (dataStr[0] == 0x0a) {
                pa++;
            } else {

                if (strlen(dataStr) > 9) {
                    printf("\nNumber too big\n");
                } else {

                    if (sscanf(dataStr, "%x", &d) < 1) {
                        /* printf("Argument missing or wrong\n"); */
                        busy = 0;
                    } else {
                        /* printf("\nWriting %08X to addr %08X\n", d, pa); */
                        *va = d;
                        pa++;
                    }
                }
            }
        }
#ifndef TEST_MODE
        munmap(map_base, MAP_SIZE);
#endif
    } while (busy);

}

int main(int argc, char** argv)
{
    char rawCmdstr[CMD_LEN_MAX];
    char*  pRaw;
    char** cmdstr=&pRaw;
    Cmd cmd;

	int n;

	for (n=0;n<argc;n++)
		printf("%s\n", argv[n]);
#ifndef TEST_MODE
    mem = open("/dev/mem",O_RDWR | O_SYNC);
    if ((int)mem == -1) {
        printf("Error: failed to open memory devicen");
        printf("(Newer Linux kernels need CONFIG_STRICT_DEVMEM to be disabled)");
        return -1;
    }
#endif

    do {
        printf("=>");
        rawCmdstr[0] = 0;
        pRaw = rawCmdstr;

        cmd = CMD_UNKNOWN;
        //if (fgets(rawCmdstr, CMD_LEN_MAX, stdin)!=NULL) {
            readCmdStr(rawCmdstr, CMD_LEN_MAX);
            printf("\n");
            cmd = getCmd(cmdstr);
            /* printf("remaining arguments: %s\n", *cmdstr); */
        //}

        switch(cmd) {
            case CMD_QUIT:
                printf("Bye.\n");
                break;

            case CMD_HELP:
                usage();
                break;

            case CMD_UNKNOWN:
                printf("Enter \"?\" to get some help...\n");
                break;
            case CMD_GETMEM:
                dump(*cmdstr);
                break;

            case CMD_SETMEM:
                modify(*cmdstr);
                break;

            case CMD_COL:
                if (sscanf(*cmdstr, "%x", &columns) < 1) {
                    printf("Argument missing or wrong\n");
                }
                printf("Columns set to %X\n", columns<1?1:columns);
                break;

            case CMD_BYTE:
                size = 1;
                break;

            case CMD_WORD:
                size = 2;
                break;

            case CMD_DWORD:
                size = 4;
                break;

            case CMD_LITTLE_E:
                littleEndian = 1;
                break;

            case CMD_BIG_E:
                littleEndian = 0;
                break;
        }

    } while (cmd!=CMD_QUIT);

#ifndef TEST_MODE
    close(mem);
#endif
    return 0;
}
