/*
* instructionparser.cpp
*
* This file contains static helper functions
*
* Copyright 2010 Tobias Wirtl, HBM <tobias.wirtl@hbm.com>
*
* The code contained herein is licensed under the GNU General Public
* License. You may obtain a copy of the GNU General Public License
* Version 2 or later at the following locations:
*
* http://www.opensource.org/licenses/gpl-license.html
* http://www.gnu.org/copyleft/gpl.html
*/

#include "instructionparser.h"
#include "helpers.h"
#include <iostream>
#include <fstream>
#include <iomanip>
#include <sstream>
#include <set>

using namespace std;


InstructionParser::InstructionParser(InstructionList *il)
{
    this->instructions = il;
    this->offset = 0;
}

void InstructionParser::setOffset(unsigned int offset)
{
    this->offset = offset;
}

void InstructionParser::setLineComments(bool lineComments)
{
    this->lineComments = lineComments;
}


//convert program lines to their binary derivate
void InstructionParser::programLinesToBinary()
{
    unsigned int i = 0;
    string bitcode;

    programLinesBinary.clear();

    cout<<setw(7)<<left<<"nr"<<setw(25)<<"instruction"<<setw(20)<<"mask"<<"bitcode"<<"\n";
    cout<<setw(68)<<setfill('-')<<"-"<<setfill(' ')<<"\n";
    for (i = 0; i < programLines.size(); i++)
    {
        Instruction* instruction = this->instructions->getInstructionOfInstructionStr(programLines[i]);
        if (instruction != 0)
        {
            bitcode = instruction->getBitcodeOfProgramLine(programLines[i],i+offset,&labels);
            cout<<setw(5)<<right<<i+offset<<"  "<<setw(25)<<left<<programLines[i]<<setw(20)<<instruction->getBitMask()<<bitcode<<"\n";
            this->programLinesBinary.push_back(bitcode);
        }
        else
            cout<<"Can't interpret \""<<programLines[i]<<"\" on Line "<<i+1<<"\n";
    }
}

//read all lines of the file and prepare for parsing
void InstructionParser::readProgramLines(const string &filename)
{
    char buf[101];
    char prev;
    unsigned int i;
    string temp;

    ifstream infile(filename.c_str());
    programLines.clear();
    labels.clear();
    if ( infile.is_open())
    {
        while (! infile.eof() )
        {
            temp.clear();
            infile.getline (buf,100);
            //clean line
            prev = buf[0];
            i = 0;
            while(buf[i]!='\0')
            {
                buf[i] = std::tolower(buf[i]);
                if (buf[i] == ',') buf[i] = ' ';
                //ignore comments
                if (buf[i] == '%') break;
		if (buf[i] == '\t') buf[i] = ' ';
                //generate label list
                if (buf[i] == ':')
                {
                    labels.insert(label_map::value_type(temp,programLines.size()+offset));
                    temp.clear();
                    buf[i] = prev = ' ';
                }
                //remove redundant free spaces and newlines
                if(     !(buf[i] == ' ' && prev == ' ')
                    &&  !(i == 0 && buf[i] == ' ')
                    &&  (buf[i] != '\n'))
                    temp += buf[i];
                prev = buf[i];
                i++;
            }
            //delete free space on the end
            if (temp.size() > 0 && temp.at(temp.size()-1) == ' ')
                temp = temp.substr(0,temp.size()-1);
            //ignore empty rows
            if (temp.size() > 0)
            {
                this->programLines.push_back(temp);
            }
        }
        programLinesToBinary();
    }
    else
        cout << "Can't open "<<filename<<"!\n";
}

void InstructionParser::binaryOutputToFile(const string &filename)
{
    ofstream outFile(filename.c_str(), std::ios::out|std::ios::binary);
    unsigned int i;
    unsigned int outputSize = this->programLinesBinary.size()*2;
    char output[outputSize];

    if (outFile.is_open())
    {
        for (i = 0; i < this->programLinesBinary.size(); i++)
        {

            output[i*2]     = Helpers::binStrToDecInt(this->programLinesBinary[i].substr(0,8));
            output[i*2 + 1] = Helpers::binStrToDecInt(this->programLinesBinary[i].substr(8,8));

        }
        outFile.write(output,outputSize);
        outFile.close();
    }
    else
        cout << "Can't open "<<filename<<"!";
}

void InstructionParser::hexBlobOutputToFile(const string &filename)
{
    ofstream outFile(filename.c_str());
    unsigned int i;

    if (outFile.is_open())
    {
        for (i = 0; i < this->programLinesBinary.size(); i++)
        {
            if ((i) && !(i%8))
                outFile<<",\n";
            else if (i)
                outFile<<", ";
            outFile<<Helpers::intToHexStr(Helpers::binStrToDecInt(this->programLinesBinary[i]),4);
        }
        outFile.close();
    }
    else
        cout << "Can't open "<<filename<<"!";
}


void InstructionParser::instructionOutputToFile(const string &filename)
{
    unsigned int i;
    ofstream outFile(filename.c_str());

    if (outFile.is_open())
    {
        for (i = 0; i < programLines.size(); i++)
            outFile<<programLines[i]<<'\n';
        outFile.close();
    }
    else
        cout << "Can't open "<<filename<<"!";
}




void InstructionParser::readTextBlob(const string &filename)
{
    string temp;
    stringstream strStream;

    programLinesBinary.clear();

    ifstream infile(filename.c_str());
    if ( infile.is_open())
    {
        while(infile>>temp)
            strStream<<Helpers::replace(Helpers::replace(temp,","," "),"0x","");
        while(strStream>>temp)
        {
            if (temp.length() == 4)
            {
                temp = Helpers::intToBinaryStr(Helpers::hexStrToDecInt(temp));
                while (temp.length() < 16)
                    temp='0'+temp; //XXX: inefficient
                this->programLinesBinary.push_back(temp);
            }
            else
                cout<<"Can't interpret "<<temp<<"!\n";
        }
        binaryToProgramLines();
    }
    else
        cout << "Can't open "<<filename<<"!\n";

}

//convert binary vector entries into human readable (programLines)
void InstructionParser::binaryToProgramLines()
{
    unsigned int i1, i2;
    int iTmpParam;
    Instruction *inst;
    string tmpParam, tmpParams;
    stringstream tmpInst;
    set<int> labelLines;

    this->programLines.clear();
    for (i1 = 0; i1 < this->programLinesBinary.size();i1++)
    {
        inst = this->instructions->getInstructionOfBitcodeStr(programLinesBinary[i1]);
        if (inst)
        {
            tmpInst.str("");
            tmpInst.clear();
            tmpInst<<setw(8)<<' '<<left<<setw(8);
            tmpInst<<inst->getInstructionStr();
            if (inst->getNumParams())
                tmpInst;
            tmpParams.clear();
            for (i2 = 0; i2 < inst->getNumParams(); i2++)
            {
                tmpParam = inst->getBitsOfParam(i2,programLinesBinary[i1]);
                if (i2>0)
                    tmpParams += ", ";
                else
                    tmpParams += ' ';

                iTmpParam = Helpers::binStrToDecInt(tmpParam);

                if (inst->isAbsoluteJump())
                {
                    if (iTmpParam >= offset && iTmpParam < this->programLinesBinary.size() + offset)
                    {
                        labelLines.insert(iTmpParam-offset);
                        tmpParams += "l" + Helpers::intToStr(iTmpParam);
                    }
                    else
                        tmpParams += Helpers::intToStr(iTmpParam);
                }
                else if (inst->isRelativeJump())
                {
                    //get 2 complement if needed
                    if (iTmpParam > 127)
                        iTmpParam = iTmpParam - 256;
                    labelLines.insert(i1 + iTmpParam + 1);
                    tmpParams += "l" + Helpers::intToStr(i1 + iTmpParam + 1 + offset);
                }
                else
                    tmpParams += Helpers::intToStr(iTmpParam);
            }
            tmpInst<<setw(20)<<tmpParams;
            if (this->lineComments)
                tmpInst<<'%'<<i1+offset;
            programLines.push_back(tmpInst.str());
        }
        else
            cout<<"Warning: Can't decode bitcode "<<programLinesBinary[i1]<<"!\n";
    }
    //add labels for referenced lines
    set<int>::iterator iter = labelLines.begin();
    while (iter != labelLines.end())
    {
        i1 = *iter;
        i2 = Helpers::intToStr(i1 + offset).size() + 2;
        if (i1 < programLines.size())
        {
            programLines[i1] = programLines[i1].substr(i2, programLines[i1].size() - i2);
            programLines[i1] = "l" + Helpers::intToStr(i1+offset) + ":" + programLines[i1];
        }
        iter++;
    }

}
